/**
 * @file Functions for rendering lists of slave summaries for various purposes. This includes
 * lists for the penthouse/facilities, selecting a slaves, facility leaders.
 *
 * For documentation see devNotes/slaveListing.md
 */

App.UI.SlaveList = {};

/**
 * @callback slaveTextGenerator
 * @param {App.Entity.SlaveState} slave
 * @param {number} index
 * @return {string}
 */
App.UI.SlaveList.render = function() {
	'use strict';
	let V;
	/** @type {string} */
	let passageName;
	/** @type {App.Entity.SlaveState[]} */
	let slaves;
	/** @type {boolean} */
	let slaveImagePrinted;

	// potentially can be a problem if played long enough to reach Number.MAX_SAFE_INTEGER
	let listID = Number.MIN_SAFE_INTEGER;

	return renderList;

	/**
	 * @param {number[]} indices
	 * @param {Array.<{index: number, rejects: string[]}>} rejectedSlaves
	 * @param {slaveTextGenerator} interactionLink
	 * @param {slaveTextGenerator} [postNote]
	 * @returns {string}
	 */
	function renderList(indices, rejectedSlaves, interactionLink, postNote) {
		V = State.variables;
		passageName = passage();
		slaves = V.slaves;
		V.assignTo = passageName; // would be passed to the "Assign" passage
		slaveImagePrinted = (V.seeImages === 1) && (V.seeSummaryImages === 1);

		let res = [];
		if (V.useSlaveListInPageJSNavigation === 1) {
			res.push(createQuickList(indices));
		}

		const fcs = App.Entity.facilities;

		// can't simply loop over fcs attributes, as there is the penthouse among them, which always exists
		const anyFacilityExists = fcs.brothel.established || fcs.club.established || fcs.dairy.established || fcs.farmyard.established || fcs.servantsQuarters.established || fcs.masterSuite.established || fcs.spa.established || fcs.clinic + fcs.schoolroom.established || fcs.cellblock.established || fcs.arcade.established || fcs.headGirlSuite.established;

		let showTransfers = false;
		if (anyFacilityExists) {
			if (passageName === "Main" || passageName === "Head Girl Suite" || passageName === "Spa" || passageName === "Brothel" || passageName === "Club" || passageName === "Arcade" || passageName === "Clinic" || passageName === "Schoolroom" || passageName === "Dairy" || passageName === "Farmyard" || passageName === "Servants' Quarters" || passageName === "Master Suite" || passageName === "Cellblock") {
				V.returnTo = passageName;
				showTransfers = true;
			}
		}

		for (const _si of indices) {
			let ss = renderSlave(_si, interactionLink, showTransfers, postNote);
			res.push(`<div id="slave-${slaves[_si].ID}" style="clear:both">${ss}</div>`);
		}

		for (const rs of rejectedSlaves) {
			const slave = slaves[rs.index];
			const rejects = rs.rejects;
			const slaveName = SlaveFullName(slave);
			let rejectString = rejects.length === 1 ?
				rejects[0] :
				`${slaveName}: <ul>${rejects.map(e => `<li>${e}</li>`).join('')}</ul>`;
			res.push(`<div id="slave_${slave.ID}" style="clear:both">${rejectString}</div>`);
		}

		$(document).one(':passagedisplay', function() {
			$('[data-quick-index]').unbind().click(function() {
				let which = this.attributes["data-quick-index"].value;
				let quick = $("div#list_index" + which);
				quick.toggleClass("hidden");
			});
			quickListBuildLinks();
		});

		if (V.useSlaveListInPageJSNavigation === 1) {
			listID++;
		}

		return res.join("");
	}


	/**
	 * @param {number} index
	 * @param {slaveTextGenerator} interactionLink
	 * @param {boolean} showTransfers
	 * @param {slaveTextGenerator} [postNote]
	 * @returns {string}
	 */
	function renderSlave(index, interactionLink, showTransfers, postNote) {
		let res = [];
		const slave = slaves[index];

		res.push(dividerAndImage(slave));
		res.push(interactionLink(slave, index));

		if ((slave.choosesOwnClothes === 1) && (slave.clothes === "choosing her own clothes")) {
			const _oldDevotion = slave.devotion;
			saChoosesOwnClothes(slave);
			slave.devotion = _oldDevotion; /* restore devotion value so repeatedly changing clothes isn't an exploit */
		}

		SlaveStatClamp(slave);
		slave.trust = Math.trunc(slave.trust);
		slave.devotion = Math.trunc(slave.devotion);
		slave.health = Math.trunc(slave.health);
		res.push(' will ');
		if ((slave.assignment === "rest") && (slave.health >= -20)) {
			res.push(`<strong><u><span class="lawngreen">rest</span></u></strong>`);
		} else if ((slave.assignment === "stay confined") && ((slave.devotion > 20) || ((slave.trust < -20) && (slave.devotion >= -20)) || ((slave.trust < -50) && (slave.devotion >= -50)))) {
			res.push(`<strong><u><span class="lawngreen">stay confined.</span></u></strong>`);
			if (slave.sentence > 0) {
				res.push(` (${slave.sentence} weeks)`);
			}
		} else if (slave.choosesOwnAssignment === 1) {
			res.push('choose her own job');
		} else {
			res.push(slave.assignment);
			if (slave.sentence > 0) {
				res.push(` ${slave.sentence} weeks`);
			}
		}
		res.push('. ');

		if ((V.displayAssignments === 1) && (passageName === "Main") && (slave.ID !== V.HeadGirl.ID) && (slave.ID !== V.Recruiter.ID) && (slave.ID !== V.Bodyguard.ID)) {
			res.push(App.UI.jobLinks.assignments(index, "Main"));
		}
		if (showTransfers) {
			res.push('<br>Transfer to: ' + App.UI.jobLinks.transfers(index));
		}
		res.push('<br/>');

		if (slaveImagePrinted) {
			res.push('&nbsp;&nbsp;&nbsp;&nbsp;');
		}

		clearSummaryCache();
		res.push(SlaveSummary(slave));

		if (postNote !== undefined) {
			res.push('<br>');
			res.push(postNote(slave, index));
		}

		return res.join('');
	}

	/**
	 * @param {number[]} indices
	 * @return {string}
	*/
	function createQuickList(indices) {
		let res = "";

		/* Useful for finding weird combinations — usages of this passage that don't yet generate the quick index.
		*	<<print 'pass/count/indexed/flag::[' + passageName + '/' + _Count + '/' + _indexed + '/' + $SlaveSummaryFiler + ']'>>
		*/

		if (((indices.length > 1) && ((passageName === "Main") && ((V.useSlaveSummaryTabs === 0) || (V.slaveAssignmentTab === "all"))))) {
			const _buttons = [];
			let _offset = -50;
			if (/Select/i.test(passageName)) {
				_offset = -25;
			}
			res += "<br />";
			/*
			 * we want <button data-quick-index="<<= listID>>">...
			 */
			const _buttonAttributes = {
				'id':`quick-list-toggle${listID}`,
				'data-quick-index': listID
			};
			res += App.UI.htag("Quick Index", _buttonAttributes, 'button');
			/*
			 * we want <div id="list_index3" class=" hidden">...
			 */
			let listIndexContent = "";

			for (const _ssii of indices) {
				const _IndexSlave = slaves[_ssii];
				const _indexSlaveName = SlaveFullName(_IndexSlave);
				const _devotionClass = getSlaveDevotionClass(_IndexSlave);
				const _trustClass = getSlaveTrustClass(_IndexSlave);
				_buttons.push({
					"data-name": _indexSlaveName,
					"data-scroll-to": `#slave-${_IndexSlave.ID}`,
					"data-scroll-offset": _offset,
					"data-devotion": _IndexSlave.devotion,
					"data-trust": _IndexSlave.trust,
					"class": `${_devotionClass} ${_trustClass}`
				});
			}
			if (_buttons.length > 0) {
				V.sortQuickList = V.sortQuickList || 'Devotion';
				listIndexContent += `//Sorting:// ''<span id="qlSort">$sortQuickList</span>.'' `;
				listIndexContent += '<<link "Sort by Devotion">>' +
					'<<set $sortQuickList = "Devotion" >>' +
					'<<replace "#qlSort">> $sortQuickList <</replace>>' +
					'<<run' + '$("#qlWrapper").removeClass("trust").addClass("devotion");' + 'sortButtonsByDevotion();' + '>>' +
					'<</link>> | ' +
					'<<link "Sort by Trust">>' +
					'<<set $sortQuickList = "Trust">>' +
					'<<replace "#qlSort">> $sortQuickList <</replace>>' +
					'<<run' + '$("#qlWrapper").removeClass("devotion").addClass("trust");' + 'sortButtonsByTrust();' + '>>' +
					'<</link>>' +
					'<br/>';
				listIndexContent += '<div id="qlWrapper" class="quicklist devotion">';
				for (const _button of _buttons) {
					const _buttonSlaveName = _button['data-name'];
					listIndexContent += App.UI.htag(_buttonSlaveName, _button, 'button');
				}
				listIndexContent += '</div>';
			}
			res += App.UI.htag(listIndexContent, {
				id: `list_index${listID}`,
				class: 'hidden'
			});
		}
		return res;
	}

	function SlaveArt(slave, option) {
		return `<<SlaveArtById ${slave.ID} ${option}>>`;
	}

	function slaveImage(s) {
		return `<div class="imageRef smlImg">${SlaveArt(s, 1)}</div>`;
	}

	function dividerAndImage(s, showImage) {
		showImage = showImage || true;
		const r = [V.lineSeparations === 0 ? "<br>" : "<hr style=\"margin:0\">"];
		if (showImage && (V.seeImages === 1) && (V.seeSummaryImages === 1)) {
			r.push(slaveImage(s));
		}
		return r.join("");
	}
}();

App.UI.SlaveList.Decoration = {};
/**
 * returns "HG", "BG", "PA", and "RC" prefixes
 * @param {App.Entity.SlaveState} slave
 * @returns {string}
 */
App.UI.SlaveList.Decoration.penthousePositions = (slave) => {
	if (App.Data.Facilities.headGirlSuite.manager.assignment === slave.assignment) {
		return '<strong><span class="lightcoral">HG</span></strong>';
	}
	if (App.Data.Facilities.penthouse.manager.assignment === slave.assignment) {
		return '<strong><span class="lightcoral">RC</span></strong>';
	}
	if (App.Data.Facilities.armory.manager.assignment === slave.assignment) {
		return '<strong><span class="lightcoral">BG</span></strong>';
	}
	if (Array.isArray(State.variables.personalAttention) && State.variables.personalAttention.findIndex(s => s.ID === slave.ID) !== -1) {
		return '<strong><span class="lightcoral">PA</span></strong>';
	}
	return '';
};

App.UI.SlaveList.SlaveInteract = {};

/**
 * @param {App.Entity.SlaveState} slave
 * @param {number} index
 * @return {string}
 */
App.UI.SlaveList.SlaveInteract.stdInteract = (slave, index) =>
	App.UI.passageLink(SlaveFullName(slave), 'Slave Interact', `$activeSlave = $slaves[${index}]`);


/**
 * @param {App.Entity.SlaveState} slave
 * @param {number} index
 * @return {string}
 */
App.UI.SlaveList.SlaveInteract.penthouseInteract = (slave, index) => {
	return App.UI.SlaveList.Decoration.penthousePositions(slave) + ' ' + App.UI.SlaveList.SlaveInteract.stdInteract(slave, index);
};

/**
 * @param {App.Entity.SlaveState} slave
 * @return {string}
 */
App.UI.SlaveList.SlaveInteract.personalAttention = (slave) =>
	App.UI.passageLink(SlaveFullName(slave), undefined, `App.UI.selectSlaveForPersonalAttention(${slave.ID});`);

/**
 * @param {App.Entity.SlaveState} slave
 * @param {number} index
 * @return {string}
 */
App.UI.SlaveList.SlaveInteract.subordinateTarget = (slave, index) =>
	App.UI.passageLink(SlaveFullName(slave), "Subordinate Targeting", `$activeSlave.subTarget = $slaves[${index}].ID`);

/**
 * @param {App.Entity.SlaveState} slave
 * @param {number} index
 * @return {string}
 */
App.UI.SlaveList.SlaveInteract.assign = (slave, index) =>
	App.UI.passageLink(SlaveFullName(slave), "Assign", `$i = ${index}`);

/**
 * @param {App.Entity.SlaveState} slave
 * @param {number} index
 * @return {string}
 */
App.UI.SlaveList.SlaveInteract.retrieve = (slave, index) =>
	App.UI.passageLink(SlaveFullName(slave), "Retrieve", `$i = ${index}`);

/**
 * Adds/removes a slave with the given id to/from the personal attention array
 * @param {number} id slave id
 */
App.UI.selectSlaveForPersonalAttention = function(id) {
	const V = State.variables;

	if (!Array.isArray(V.personalAttention)) {
		/* first PA target */
		V.personalAttention = [{
			ID: id,
			trainingRegimen: "undecided"
		}];
	} else {
		const _pai = V.personalAttention.findIndex(function(s) {
			return s.ID === id;
		});
		if (_pai === -1) {
			/* not already a PA target; add */
			V.activeSlave = getSlave(id);
			V.personalAttention.push({
				ID: id,
				trainingRegimen: "undecided"
			});
		} else {
			/* already a PA target; remove */
			V.personalAttention.deleteAt(_pai);
			if (V.personalAttention.length === 0) {
				V.personalAttention = "sex";
			}
		}
	}
	SugarCube.Engine.play("Personal Attention Select");
};

/**
 * Generates fragment with sorting options, that link to the given pasage
 * @param {string} passage The passage to link to
 * @returns {string}
 */
App.UI.SlaveList.sortingLinks = function (passage) {
	const V = State.variables;
	let r = '&nbsp;&nbsp;&nbsp;&nbsp;Sort by: ';
	r += ["devotion", "name", "assignment", "seniority", "actualAge", "visualAge", "physicalAge"]
		.map(so => V.sortSlavesBy !== so ?
			App.UI.passageLink(capFirstChar(so.replace(/([A-Z])/g, " $1")), passage, `$sortSlavesBy = "${so}"`) : capFirstChar(so))
		.join("&thinsp;|&thinsp;");

	r += '&nbsp;&nbsp;&nbsp;&nbsp;Sort: ';
	r += ["descending", "ascending"].map(so => V.sortSlavesOrder !== so ?
		App.UI.passageLink(capFirstChar(so), passage, `$sortSlavesOrder = "${so}"`) : capFirstChar(so))
		.join("&thinsp;|&thinsp;");
	return r;
};

/**
 * Standard tabs for a facility with a single job (SJ)
 * @param {App.Entity.Facilities.Facility} facility
 * @param {string} facilityPassage
 * @param {boolean} [showTransfersTab=false]
 * @param {{assign: string, remove: string, transfer: (string| undefined)}} [tabCaptions]
 * @returns {string}
 */
App.UI.SlaveList.listSJFacilitySlaves = function (facility, facilityPassage, showTransfersTab = false, tabCaptions = undefined) {
	const V = State.variables;
	tabCaptions = tabCaptions || {
		assign: 'Assign a slave',
		remove: 'Remove a slave',
		transfer: 'Transfer from Facility'
	};
	let r = '';
	if (V.sortSlavesMain) {
		r += this.sortingLinks(facilityPassage) + '<br>';
	}
	r += '<div class="tab">' +
		App.UI.tabbar.tabButton('assign',  tabCaptions.assign) +
		App.UI.tabbar.tabButton('remove', tabCaptions.remove) +
		(showTransfersTab ? App.UI.tabbar.tabButton('transfer', tabCaptions.transfer) : '')+
	'</div>';

	if (facility.hostedSlaves > 0) {
		let facilitySlaves = facility.job().employeesIndices();
		SlaveSort.indices(facilitySlaves);
		r += App.UI.tabbar.makeTab("remove", App.UI.SlaveList.render(facilitySlaves, [],
			App.UI.SlaveList.SlaveInteract.stdInteract,
			(slave, index) => App.UI.passageLink(`Retrieve ${slave.object} from ${facility.name}`, "Retrieve", `$i = ${index}`)));
	} else {
		r += App.UI.tabbar.makeTab("remove", `<em>${capFirstChar(facility.name)} is empty for the moment</em>`);
	}

	/**
	 * @param {number[]} slaveIdxs
	 * @returns {string}
	 */
	function assignableTabContent(slaveIdxs) {
		SlaveSort.indices(slaveIdxs);
		const slaves = V.slaves;
		let rejectedSlaves = [];
		let passedSlaves = [];
		slaveIdxs.forEach((idx) => {
			const rejects = facility.canHostSlave(slaves[idx]);
			if (rejects.length > 0) {
				rejectedSlaves.push({index: idx, rejects: rejects});
			} else {
				passedSlaves.push(idx);
			}
		}, []);
		return App.UI.SlaveList.render(passedSlaves, rejectedSlaves,
			App.UI.SlaveList.SlaveInteract.stdInteract,
			(slave, index) => App.UI.passageLink(`Send ${slave.object} to ${facility.name}`, "Assign", `$i = ${index}`));
	}
	if (facility.hasFreeSpace) {
		// slaves from the penthouse can be transferred here
		r += App.UI.tabbar.makeTab("assign", assignableTabContent(App.Entity.facilities.penthouse.employeesIndices()));
	} else {
		r += App.UI.tabbar.makeTab("assign", `<strong>${capFirstChar(facility.name)} is full and cannot hold any more slaves</strong>`);
	}

	if (showTransfersTab) {
		if (facility.hasFreeSpace) {
			// slaves from other facilities can be transferred here
			const transferableIndices = V.slaves.reduce((acc, slave, ind) => {
				if (slave.assignmentVisible === 0 && !facility.isHosted(slave)) {
					acc.push(ind);
				}
				return acc;
			}, []);
			r += App.UI.tabbar.makeTab("transfer", assignableTabContent(transferableIndices));
		} else {
			r += App.UI.tabbar.makeTab("transfer", `<strong>${capFirstChar(facility.name)} is full and cannot hold any more slaves</strong>`);
		}
	}
	App.UI.tabbar.handlePreSelectedTab();

	return r;
};

/**
 * @returns {string}
 */
App.UI.SlaveList.listNGPSlaves = function () {
	const V = State.variables;
	const thisPassage = 'New Game Plus';
	let r = this.sortingLinks(thisPassage) + '<br>';

	r += '<div class="tab">' +
		App.UI.tabbar.tabButton('assign', 'Import a slave') +
		App.UI.tabbar.tabButton('remove', 'Remove from impor') +
	'</div>';

	const NGPassignment = "be imported";
	/** @type {App.Entity.SlaveState[]} */
	const slaves = V.slaves;

	if (V.slavesToImport > 0) {
		const importedSlavesIndices = slaves.reduce((acc, s, i) => { if (s.assignment === NGPassignment) { acc.push(i); } return acc; }, []);
		SlaveSort.indices(importedSlavesIndices);
		r += App.UI.tabbar.makeTab("remove", App.UI.SlaveList.render(importedSlavesIndices, [],
			(s) => `<u><strong><span class="pink">${SlaveFullName(s)}</span></strong></u>`,
			(s) => App.UI.passageLink('Remove from import list', thisPassage, `$slavesToImport -= 1, removeJob(${s}, "${NGPassignment}")`)));
	} else {
		r += App.UI.tabbar.makeTab("remove", `<em>No slaves will go with you to the new game</em>`);
	}

	if (V.slavesToImport < V.slavesToImportMax) {
		const slavesToImportIndices = slaves.reduce((acc, s, i) => { if (s.assignment !== NGPassignment) { acc.push(i); } return acc; }, []);
		SlaveSort.indices(slavesToImportIndices);
		r += App.UI.tabbar.makeTab("assign", App.UI.SlaveList.render(slavesToImportIndices, [],
			(s) => `<u><strong><span class="pink">${SlaveFullName(s)}</span></strong></u>`,
			(s) => App.UI.passageLink('Add to import list', thisPassage, `$slavesToImport += 1, assignJob(${s}, "${NGPassignment}")`)));
	} else {
		r += App.UI.tabbar.makeTab("assign", `<strong>Slave import limit reached</strong>`);
	}

	App.UI.tabbar.handlePreSelectedTab();
	return r;
};

/**
 * Renders facility manager summary or a note with a link to select one
 * @param {App.Entity.Facilities.Facility} facility
 * @param {string} [selectionPassage] passage name for manager selection. "${Manager} Select" if omitted
 * @returns {string}
 */
App.UI.SlaveList.displayManager = function (facility, selectionPassage) {
	const managerCapName = capFirstChar(facility.desc.manager.position);
	selectionPassage = selectionPassage || `${managerCapName} Select`;
	const manager = facility.manager.currentEmployee;
	if (manager) {
		return this.render([App.Utils.slaveIndexForId(manager.ID)], [],
			App.UI.SlaveList.SlaveInteract.stdInteract,
			() => App.UI.passageLink(`Change or remove ${managerCapName}`, selectionPassage, ""));
	} else {
		return `You do not have a slave serving as a ${managerCapName}. ${App.UI.passageLink(`Appoint one`, selectionPassage, "")}`;
	}
};

/**
 * Displays standard facility page with manager and list of workers
 * @param {App.Entity.Facilities.Facility} facility
 * @param {boolean} [showTransfersPage]
 * @returns {string}
 */
App.UI.SlaveList.stdFacilityPage = function (facility, showTransfersPage) {
	return this.displayManager(facility) + '<br><br>' + this.listSJFacilitySlaves(facility, passage(), showTransfersPage);
};

App.UI.SlaveList.penthousePage = function () {
	const V = State.variables;
	const ph = App.Entity.facilities.penthouse;
	const listElementId = 'summarylist'; // for the untabbed mode only

	function span(text, cls, id) {
		return `<span${cls ? ` class="${cls}"` : ''}${id ? ` id="${id}"` : ''}>${text}</span>`;
	}

	function overviewTabContent() {
		let r = '';
		const thisArcology = V.arcologies[0];

		/** @type {App.Entity.SlaveState} */
		const HG = V.HeadGirl;
		if (HG) {
			r += `<strong><u>${span(SlaveFullName(HG), "pink")}</u></strong> is serving as your Head Girl`;
			if (thisArcology.FSEgyptianRevivalistLaw === 1) {
				r += ' and Consort';
			}
			r += `. <strong> ${span(App.UI.passageLink("Manage Head Girl", "HG Select"), null, "manageHG")}</strong>  ${span("[H]", "cyan")}`;
			r += App.UI.SlaveList.render([App.Utils.slaveIndexForId(HG.ID)], [],
				App.UI.SlaveList.SlaveInteract.penthouseInteract);
		} else {
			if (V.slaves.length > 1) {
				r += `You have ${span("not", "red")} selected a Head Girl`;
				if (thisArcology.FSEgyptianRevivalistLaw === 1) {
					r += ' and Consort';
				}
				r += `. <strong>${span(App.UI.passageLink("Select One", "HG Select"), null, "manageHG")}</strong> ${span("[H]", "cyan")}`;
			} else {
				r += '<em>You do not have enough slaves to keep a Head Girl</em>';
			}
		}
		r += '<br>';

		/** @type {App.Entity.SlaveState} */
		const RC = V.Recruiter;
		if (RC) {
			const p = getPronouns(RC);
			r += `<strong><u>${span(SlaveFullName(RC), "pink")}</u></strong> is working `;
			if (V.recruiterTarget !== "other arcologies") {
				r += 'to recruit girls.';
			} else {
				r += 'as a Sexual Ambassador';
				if (thisArcology.influenceTarget === -1) {
					r += ', but ' + span(p.object + ' has no target to influence', "red");
				} else {
					const targetName = V.arcologies.find(a => a.direction === thisArcology.influenceTarget).name;
					r += ' to ' + targetName + '.';
				}
			}
			r += `${span('<strong>' + App.UI.passageLink("Manage Recruiter", "Recruiter Select") + '</strong>', null, "manageRecruiter")} ${span("[U]", "cyan")}`;
			r += App.UI.SlaveList.render([App.Utils.slaveIndexForId(RC.ID)], [],
				App.UI.SlaveList.SlaveInteract.penthouseInteract);
		} else {
			r += `You have ${span("not", "red")} selected a Recruiter. `;
			r += `${span('<strong>' + App.UI.passageLink("Select one", "Recruiter Select") + '</strong>', null, "manageRecruiter")} ${span("[U]", "cyan")}`;
		}

		if (V.dojo) {
			r += '<br>';
			/** @type {App.Entity.SlaveState} */
			const BG = V.Bodyguard;
			if (BG) {
				r += `<strong><u>${span(SlaveFullName(RC), "pink")}</u></strong> is serving as your bodyguard. `;
				r += span(`<strong>${App.UI.passageLink("Manage Bodyguard", "BG Select")}</strong>`, null, "manageBG") +
					span("[B]", "cyan");
				r += App.UI.SlaveList.render([App.Utils.slaveIndexForId(BG.ID)], [],
					App.UI.SlaveList.SlaveInteract.penthouseInteract);
			} else {
				r += `You have ${span("not", "red")} selected a Bodyguard. `;
				r += span(`<strong>${App.UI.passageLink("Select one", "BG Select")}</strong>`, null, "manageBG") +
					span("[B]", "cyan");
			}

			/* Start Italic event text */
			if (BG && BG.assignment === "guard you") {
				const p = getPronouns(BG);
				V.i = App.Utils.slaveIndexForId(BG.ID);
				const interactLinkSetters = `$activeSlave = $slaves[${V.i}], $nextButton = "Back", $nextLink = "AS Dump", $returnTo = "Main"`;
				r += '<br>';
				// <<= App.Interact.UseGuard($slaves[$i])>>//
				let useHimLinks = [];
				useHimLinks.push(App.UI.passageLink(`Use ${p.his} mouth`, "Flips", interactLinkSetters));
				useHimLinks.push(App.UI.passageLink(`Play with ${p.his} tits`, "FBoobs", interactLinkSetters));
				if (canDoVaginal(BG)) {
					useHimLinks.push(App.UI.passageLink(`Fuck ${p.him}`, "FVagina", interactLinkSetters));
					if (canDoAnal(BG)) {
						useHimLinks.push(App.UI.passageLink(`Use ${p.his} holes`, "FButt", interactLinkSetters));
					}
					if (BG.belly >= 300000) {
						useHimLinks.push(App.UI.passageLink(`Fuck ${p.him} over ${p.his} belly`, "FBellyFuck", interactLinkSetters));
					}
				}
				/* check */
				if (canPenetrate(BG)) {
					useHimLinks.push(App.UI.passageLink(`Ride ${p.him}`, "FDick", interactLinkSetters));
				}
				if (canDoAnal(BG)) {
					useHimLinks.push(App.UI.passageLink(`Fuck ${p.his} ass`, "FAnus", interactLinkSetters));
				}
				useHimLinks.push(App.UI.passageLink(`Abuse ${p.him}`, "Gameover", '$gameover ="idiot ball"'));

				r += `<br>&nbsp;&nbsp;&nbsp;&nbsp;${useHimLinks.join('&thinsp;|&thinsp;')}`;
				/* End Italic event text */
			}
			r += "<br/>";
		}
		return r;
	}

	/**
	 * @param {string} job
	 * @returns {{n: number, text: string}}
	 */
	function _slavesForJob(job) {
		const employeesIndices = job === 'all' ? ph.employeesIndices() : ph.job(job).employeesIndices();
		if (employeesIndices.length === 0) { return {n: 0, text: ''}; }

		SlaveSort.indices(employeesIndices);
		return {
			n: employeesIndices.length,
			text: App.UI.SlaveList.render(employeesIndices, [], App.UI.SlaveList.SlaveInteract.penthouseInteract)
		};
	}

	/**
	 * Displays job filter links, whose action are generated by the callback
	 * @param {assignmentFilterGenerateCallback} callback
	 * @returns {string}
	 */
	function _jobFilter(callback) {
		const jd = App.Data.Facilities.penthouse.jobs;
		let links = [];
		links.push(`<<link "All">>${callback('all')}<</link>>`);
		// seems like SC2 does not process data-setter when data-passage is not set
		for (const jn in jd) {
			links.push(`<<link "${capFirstChar(jd[jn].position)}">>${callback(jn)}<</link>>`);
		}

		return links.join('&thinsp;|&thinsp;');
	}

	function _updateList(job) {
		State.temporary.mainPageUpdate.job = job;
		App.UI.replace('#' + listElementId, _slavesForJob(job).text);
	}

	/**
	 * @typedef tabDesc
	 * @property {string} tabName
	 * @property {string} caption
	 * @property {string} content
	 */

	/**
	 * @param {string} tabName
	 * @param {string} caption
	 * @param {string} content
	 * @returns {tabDesc}
	 */
	function makeTabDesc(tabName, caption, content) {
		return {
			tabName: tabName,
			caption: caption,
			content: content
		};
	}

	let r = '';

	if (V.positionMainLinks >= 0) {
		r += '<center>' + App.UI.View.MainLinks() + '</center><br>';
	}

	if (V.sortSlavesMain) {
		r += '<br>' + this.sortingLinks("Main") + '<br>';
	}

	if (V.useSlaveSummaryTabs) {
		/** @type {tabDesc[]} */
		let tabs = [];

		if (V.useSlaveSummaryOverviewTab) {
			tabs.push(makeTabDesc('overview', 'Overview', overviewTabContent()));
		}

		for (const jn of ph.jobsNames) {
			const slaves = _slavesForJob(jn);
			if (slaves.n > 0) {
				tabs.push(makeTabDesc(jn, `${ph.desc.jobs[jn].position} (${slaves.n})`, slaves.text));
			}
		}

		// now generate the "All" tab
		const penthouseSlavesIndices = ph.employeesIndices();
		SlaveSort.indices(penthouseSlavesIndices);
		tabs.push(makeTabDesc('all', `All (${penthouseSlavesIndices.length})`,
			this.render(penthouseSlavesIndices, [], App.UI.SlaveList.SlaveInteract.penthouseInteract)));


		r += '<div class="tab">';
		for (const tab of tabs) {
			r += App.UI.tabbar.tabButton(tab.tabName, tab.caption);
		}
		r += '</div>';

		for (const tab of tabs) {
			r += App.UI.tabbar.makeTab(tab.tabName, tab.content);
		}
	} else {
		State.temporary.mainPageUpdate = {
			job: State.temporary.mainPageUpdate ? State.temporary.mainPageUpdate.job : 'all',
			update: _updateList
		};

		$(document).one(':passagedisplay', () => { _updateList(State.temporary.mainPageUpdate.job); });
		r += `<div>${_jobFilter(s => `<<run _mainPageUpdate.update("${s}")>>`)} <div id=${listElementId}></div></div>`;
	}

	if (V.positionMainLinks <= 0) {
		r += '<br><center>' + App.UI.View.MainLinks() + '</center>';
	}

	App.UI.tabbar.handlePreSelectedTab();
	return r;
};

/**
 * @callback assignmentFilterGenerateCallback
 * @param {string} value
 * @returns {string}
 */

/**
 * @callback slaveFilterCallbackReasoned
 * @param {App.Entity.SlaveState} slave
 * @returns {string[]}
 */

 /**
 * @callback slaveFilterCallbackSimple
 * @param {App.Entity.SlaveState} slave
 * @returns {boolean}
 */

 /**
 * @callback slaveTestCallback
 * @param {App.Entity.SlaveState} slave
 * @returns {boolean}
 */

App.UI.SlaveList.slaveSelectionList = function () {
	const selectionElementId = "slaveSelectionList";

	return selection;

	/**
	 * @typedef ListOptions
	 * @property {slaveFilterCallbackReasoned|slaveFilterCallbackSimple} filter
	 * @property {slaveTestCallback} [expCheck]
	 * @property {slaveTextGenerator} interactionLink
	 * @property {slaveTextGenerator} [postNote]
	 */

	/**
	 * @param {slaveFilterCallbackReasoned|slaveFilterCallbackSimple} filter
	 * @param {slaveTextGenerator} interactionLink
	 * @param {slaveTestCallback} [experianceChecker]
	 * @param {slaveTextGenerator} [postNote]
	 * @returns {string}
	*/
	function selection(filter, interactionLink, experianceChecker, postNote) {
		if (experianceChecker === null) { experianceChecker = undefined; }
		State.temporary.slaveSelection = {
			filter: filter,
			expCheck: experianceChecker,
			interactionLink: interactionLink,
			postNote: postNote,
			update: _updateList
		};

		$(document).one(':passagedisplay', () => { _updateList('all'); });
		return `<div>${_assignmentFilter(s => `<<run _slaveSelection.update('${s}')>>`, experianceChecker !== undefined)} <div id=${selectionElementId}></div></div>`;
	}

	function _updateList(assignment) {
		App.UI.replace('#' + selectionElementId, _listSlaves(assignment, State.temporary.slaveSelection));
	}
	/**
	 * Displays assignment filter links, whose action are generated by the callback
	 * @param {assignmentFilterGenerateCallback} callback
	 * @param {boolean} includeExperienced
	 * @returns {string}
	 */
	function _assignmentFilter(callback, includeExperienced) {
		let filters = {
			all: "All"
		};
		let fNames = Object.keys(App.Entity.facilities);
		fNames.sort();
		for (const fn of fNames) {
			/** @type {App.Entity.Facilities.Facility} */
			const f = App.Entity.facilities[fn];
			if (f.established && f.hostedSlaves > 0) {
				filters[fn] = f.name;
			}
		}
		let links = [];
		/* seems like SC2 does not process data-setter when data-passage is not set
		for (const f in filters) {
			links.push(App.UI.passageLink(filters[f], passage, callback(f)));
		}
		if (includeExperienced) {
			links.push(`<span class="lime">${App.UI.passageLink('Experienced', passage, callback('experienced'))}</span>`);
		}*/
		for (const f in filters) {
			links.push(`<<link "${filters[f]}">>${callback(f)}<</link>>`);
		}
		if (includeExperienced) {
			links.push(`<span class="lime"><<link "Experienced">>${callback('experienced')}<</link>></span>`);
		}

		return links.join('&thinsp;|&thinsp;');
	}

	/**
	 *
	 * @param {string} assignmentStr
	 * @param {ListOptions} options
	 * @returns {string}
	 */
	function _listSlaves(assignmentStr, options) {
		/** @type {App.Entity.SlaveState[]} */
		const slaves = State.variables.slaves;
		let unfilteredIndices = [];
		switch (assignmentStr) {
			case 'all':
				unfilteredIndices = Array.from({length: slaves.length}, (v, i) => i);
				break;
			case 'experienced':
				unfilteredIndices = slaves.reduce((acc, s, idx) => {
					if (options.expCheck(s)) {
						acc.push(idx);
					}
					return acc;
				}, []);
				break;
			default:
				unfilteredIndices = App.Entity.facilities[assignmentStr].employeesIndices();
				break;
		}
		SlaveSort.indices(unfilteredIndices);
		let passingIndices = [];
		let rejects = [];

		unfilteredIndices.forEach(idx => {
			const fr = options.filter(slaves[idx]);
			if (fr === true || (Array.isArray(fr) && fr.length === 0)) {
				passingIndices.push(idx);
			} else {
				if (Array.isArray(fr)) { rejects.push({index: idx, rejects: fr}); }
			}
		});

		// clamsi fragment to create a function which combines results of two optional tests
		// done this way to test for tests presense only once
		const listPostNote = options.expCheck ?
			(options.postNote ?
				(s, i) => options.expCheck(s) ? '<span class="lime">Has applicable career experience.</span><br>' : '' + options.postNote(s, i) :
				(s) => options.expCheck(s) ? '<span class="lime">Has applicable career experience.</span>' : '') :
			options.postNote ?
				(s, i) => options.postNote(s, i) :
				() => '';

		return App.UI.SlaveList.render(passingIndices, rejects, options.interactionLink, listPostNote);
	}
}();

/**
 * @param {App.Entity.Facilities.Facility} facility
 * @param {string} [passage] one of the *Workaround passages. Will be composed from the position name if omitted
 * @returns {string}
 */
App.UI.SlaveList.facilityManagerSelection = function (facility, passage) {
	passage = passage || capFirstChar(facility.manager.desc.position) + " Workaround";
	return this.slaveSelectionList(slave => facility.manager.canEmploy(slave),
		(slave, index) => App.UI.passageLink(SlaveFullName(slave), passage, `$i = ${index}`),
		slave => facility.manager.slaveHasExperience(slave));
};
